#include "osd_bb.h"
#include "linux_dbg.h"

/*
TO_DO_LIST:

*/
#ifdef OSD_BOUNCE_BUFFER
int bb_ret;

void bb_msg(int level, char *fmt,...){
	va_list ap;
	char buf[STR_SZ];

	va_start(ap, fmt);

	vsprintf(buf, fmt, ap);
	switch( level ){
	case BB_ERROR:
	case BB_INFO:
	case BB_WARN:
		printk(buf);
		break;
	case BB_DEBUG:
	case BB_MSG:
	default:
		DO_NOTHING;
	}
	
	va_end(ap);
}

/*
 reset object's values. This is REQUIRED.
*/
void reset_sg_translator(pSGTRANS_T obj){
	int i;

        obj->cmd = NULL;
        obj->old_sgcnt = 0;
        obj->old_req_buf = NULL;
        obj->new_req_buf = obj->request_buffer;
        for(i = 0; i<BB_MAX_SG; i++){
	        obj->request_buffer[i].length = BB_PAGE_SIZE;
	}

}

pSGTRANS_T get_sg_translator(bb_t *bb){
        pSGTRANS_T obj = NULL;
        unsigned long flags;

        spin_lock_irqsave(&(bb->sg_trans_lock), flags);
        if( bb->pSgTransCurrent ){
                obj = bb->pSgTransCurrent;
                bb->pSgTransCurrent = obj->next;
                obj->next = NULL;
        }
	spin_unlock_irqrestore(&bb->sg_trans_lock, flags);

	if( !obj )
		bb_msg(BB_ERROR, "get_sg_translator failed\n");
	
	//__flush_tlb_global();
	return obj;
}

int put_sg_translator(bpOSD_COMP_EXTENSION p, pSGTRANS_T obj){
        unsigned long flags;
	bb_t	*bb = &p->bb;

        spin_lock_irqsave(&bb->sg_trans_lock, flags);

	reset_sg_translator(obj);
        obj->next = bb->pSgTransCurrent;
        bb->pSgTransCurrent = obj;

        spin_unlock_irqrestore(&bb->sg_trans_lock, flags);

        return 0;
}

#define IS_SG   0
#define IS_PG   1
unsigned char *get_vaddr( unsigned char *obj, int obj_type){
	struct scatterlist 	*sg;
	unsigned char 		*vaddr;

	if( obj_type == IS_SG ){
		sg = (struct scatterlist*) obj;
		
		vaddr = kmap_atomic(t3_sg_page(sg), KM_IRQ0);
		vaddr += sg->offset;

		bb_msg(BB_DEBUG, "vaddr = %p\n", vaddr);
	}else{	// type==IS_PG
		vaddr = kmap_atomic( (struct page*)obj, KM_IRQ0);
	}
	
	return vaddr;
}

void free_vaddr(unsigned char *vaddr){

	kunmap_atomic( (unsigned long)vaddr & PAGE_MASK, KM_IRQ0 );
}


int	handled_within_a_pg(sg_optable_t *tbl){
	int val;

	/* There is ONLY src or dst would be highmem, that is, one side host and the other side bounce buffer.	*/
	if( tbl->src_is_highmem ){
		val = tbl->src_high_remain;
	}else if( tbl->dst_is_highmem ){
		val = tbl->dst_high_remain; 
	}else{
		val = BB_PAGE_SIZE;
	}

	return val;
}

#define IS_HIGHPAGE(x)	PageHighMem(t3_sg_page(x))
#define GET_HIGHPAGE(x) (t3_sg_page(x))
#define PAGE_REMAINDER(x)	(PAGE_SIZE - ((unsigned int)(x) & ~PAGE_MASK))
/* 
source walker:
INPUT:	operation table
RETURN:	0 for EOF
*/
int s_walker( sg_optable_t *tbl ){
	struct scatterlist *src;
	int ret=0;

	bb_msg(BB_DEBUG, "src_buff_remain=%d, handled_buff_sz=%d\n", tbl->src_buff_remain, tbl->handled_buff_sz);
	if( tbl->src_buff_remain )
		tbl->src_buff_remain -= tbl->handled_buff_sz;
	if( tbl->src_high_remain )
		tbl->src_high_remain -= tbl->handled_buff_sz;

	if( !tbl->src_buff_remain ){
		tbl->current_src_idx ++;
		if(tbl->current_src_idx < tbl->src_sgcnt){
			src = (tbl->src_sg + tbl->current_src_idx);
	
			tbl->src_is_highmem = IS_HIGHPAGE(src);

			// Preparing virtual address.
			if(tbl->src_mapped_vaddr){
				free_vaddr( tbl->src_mapped_vaddr);		// what if invalid released other process's?
				tbl->src_mapped_vaddr = 0;
			}
			JDBG;
			tbl->current_src_vaddr = get_vaddr(src, IS_SG);
			tbl->src_mapped_vaddr = tbl->current_src_vaddr;
			tbl->src_buff_remain = src->length;

			if( tbl->src_is_highmem ){
                                tbl->src_high_page = GET_HIGHPAGE(src);
                                tbl->src_high_remain = 
					min ( 	PAGE_REMAINDER(tbl->current_src_vaddr), 
						tbl->src_buff_remain );
                        }
			ret = 1;
		}else{
			if( tbl->src_mapped_vaddr ){
				free_vaddr( tbl->src_mapped_vaddr );
				tbl->src_mapped_vaddr = 0;
			}
			
			BB_ASSERT( tbl->src_high_remain==tbl->src_buff_remain);
			ret = 0;
		}
	}else{ // src repositioning
		if( !tbl->src_is_highmem )
			tbl->current_src_vaddr += tbl->handled_buff_sz;
		else{
			if( tbl->src_high_remain ){
				tbl->current_src_vaddr += tbl->handled_buff_sz;
			}else{ // move to next highpage
				free_vaddr( tbl->src_mapped_vaddr );
				tbl->src_high_page++;
				JDBG;
				tbl->current_src_vaddr = get_vaddr( tbl->src_high_page, IS_PG);
				tbl->src_mapped_vaddr = tbl->current_src_vaddr;
				tbl->src_high_remain = 
					min( 	PAGE_REMAINDER(tbl->current_src_vaddr), 
						tbl->src_buff_remain );
			}
		}
		ret = 1;
	}

	return ret;
}


int d_walker( sg_optable_t *tbl ){
	struct scatterlist *dst;
	int	ret;

	bb_msg(BB_DEBUG, "dst_buff_remain=%d, handled_buff_sz=%d\n", tbl->dst_buff_remain, tbl->handled_buff_sz);
	if( tbl->dst_buff_remain )
		tbl->dst_buff_remain -= tbl->handled_buff_sz; 	// sub last processed chunk.
	if( tbl->dst_high_remain )
                tbl->dst_high_remain -= tbl->handled_buff_sz;
	
	if( !tbl->dst_buff_remain ){
		tbl->current_dst_idx ++;
		if(tbl->current_dst_idx < tbl->dst_sgcnt){
			dst = ( tbl->dst_sg + tbl->current_dst_idx );

			tbl->dst_is_highmem = IS_HIGHPAGE(dst);

			// Prepare virtual address.
			if(tbl->dst_mapped_vaddr){
                                free_vaddr( tbl->dst_mapped_vaddr);
				tbl->dst_mapped_vaddr = 0;
			}
	                tbl->current_dst_vaddr = get_vaddr(dst, IS_SG);
			tbl->dst_mapped_vaddr = tbl->current_dst_vaddr;
			tbl->dst_buff_remain = dst->length;

			if( tbl->dst_is_highmem ){
                                tbl->dst_high_page = GET_HIGHPAGE(dst);
                                tbl->dst_high_remain = 
					min (	PAGE_REMAINDER( tbl->current_dst_vaddr ),
						tbl->dst_buff_remain );
                        }
                        ret = 1;

			dst->length = 0; // correct value will be assigned in below "else" section.
		}else{
                        if( tbl->dst_mapped_vaddr ){
                                free_vaddr( tbl->dst_mapped_vaddr );
				tbl->dst_mapped_vaddr = 0;
			}
                        ret = 0;
                }
	}else{ //dst repositioning
		dst = ( tbl->dst_sg + tbl->current_dst_idx );
		if( !tbl->dst_is_highmem )
                        tbl->current_dst_vaddr += tbl->handled_buff_sz;
                else{
                        if( tbl->dst_high_remain ){
                                tbl->current_dst_vaddr += tbl->handled_buff_sz;
                        }else{ // move to next highpage
                                tbl->dst_high_page++;
                                free_vaddr( tbl->dst_mapped_vaddr );
                                tbl->current_dst_vaddr = get_vaddr( tbl->dst_high_page, IS_PG);
				tbl->dst_mapped_vaddr = tbl->current_dst_vaddr;
                                tbl->dst_high_remain = 
					min (	PAGE_REMAINDER( tbl->current_dst_vaddr ),
						tbl->dst_buff_remain );
                        }
                }
                ret = 1;
	}
	
	// denote current chunk.
	tbl->handled_buff_sz = min( handled_within_a_pg(tbl) , min(tbl->src_buff_remain, tbl->dst_buff_remain) );
	bb_msg(BB_DEBUG, "handled_buff_sz = %d\n", tbl->handled_buff_sz);

	dst->length += tbl->handled_buff_sz;

	return 0;
}

void move_data( sg_optable_t *tbl ){

	if( tbl->r_o_w == DO_COPY ){
                memcpy(tbl->current_dst_vaddr, tbl->current_src_vaddr, tbl->handled_buff_sz);
        }
}

int prepare_sg_optbl(
		sg_optable_t *tbl, 
		struct scatterlist *src_sg, 
		struct scatterlist *dst_sg, 
		int src_sgcnt, 
		int dst_sgcnt, 
		int r_o_w)
{
                tbl->src_sg   = src_sg;
                tbl->dst_sg   = dst_sg;
                tbl->src_sgcnt = src_sgcnt;
                tbl->dst_sgcnt = dst_sgcnt;
                tbl->current_src_idx = -1;
                tbl->current_dst_idx = -1;
                tbl->r_o_w    = r_o_w;
		tbl->current_src_vaddr = NULL;
		tbl->current_dst_vaddr = NULL;
		tbl->src_buff_remain = 0;
		tbl->dst_buff_remain = 0;
		tbl->handled_buff_sz = 0;
		tbl->src_is_highmem = 0;
		tbl->src_high_remain = 0;
		tbl->src_mapped_vaddr = 0;
		tbl->dst_is_highmem = 0;
		tbl->dst_high_remain = 0;
		tbl->dst_mapped_vaddr = 0;
}

void walk_start(sg_optable_t *tbl){
	DO_NOTHING;
}

void walk_end(sg_optable_t *tbl){
	if( tbl->src_mapped_vaddr )
		free_vaddr( tbl->src_mapped_vaddr );

	if( tbl->dst_mapped_vaddr )
		free_vaddr( tbl->dst_mapped_vaddr );
}

/* Translate multiple sg levels to BB_MAX_SG level. Currently, BB_MAX_SG == 1	*/
int sg_copy(	struct scatterlist *src_sg, struct scatterlist *dst_sg, 
		int src_sgcnt, int dst_sgcnt, 
		int r_o_w)
{
	sg_optable_t    walk_tbl;
	int	use_sg;

	prepare_sg_optbl( &walk_tbl, src_sg, dst_sg, src_sgcnt, dst_sgcnt, r_o_w);

	walk_start(&walk_tbl);
	while( s_walker(&walk_tbl) ){
		d_walker(&walk_tbl);
		move_data(&walk_tbl);
	}
	walk_end(&walk_tbl);

	use_sg = walk_tbl.current_dst_idx + 1;
	return use_sg;	
}

int live_or_die(Scsi_Cmnd* cmd, bpOSD_COMP_EXTENSION p){
	if( !p->bb.present )
		return BB_DIE;

        if( !(READ_CMD(cmd) || WRITE_CMD(cmd)) )
                return BB_DIE;

	#if 0
	if( !Engine_IsArrayCritical( (bpRaidEngine)p->ComponentResource[COMP_TYPE_ENGINE].ComponentExtension, cmd->device->id)  )
		return BB_DIE;
	#endif

	if( !t3_use_sg(cmd) )
		return BB_DIE;

	return BB_ALIVE;
}

int sg2bb(Scsi_Cmnd* cmd, bpOSD_COMP_EXTENSION p){
	bb_t	*bb = &p->bb;
        pSGTRANS_T     	pSgTrans;
	struct scatterlist      *new_sg, *old_sg;
	unsigned long 	flags;

	if( live_or_die(cmd, p) == BB_DIE)
		return -1;

	if(bb->cmd_reject_threshold >= BB_CAN_QUEUE)
                return NOBBPAGE;

	bb->cmd_reject_threshold++;

	bb_msg(BB_DEBUG, "sg2bb CMD: %p, %x\n", cmd, cmd->cmnd[0]);

        /* get a SgTrans obj    */
        if((pSgTrans=get_sg_translator(bb)) == NULL){
		bb_msg(BB_ERROR, "get_sg_translator failed\n");
		return -BBERR_NOMEM;
	}

	pSgTrans->cmd = cmd;
	pSgTrans->old_sgcnt = t3_use_sg(cmd);

	pSgTrans->old_req_buf = (unsigned char *)t3_request_buffer(cmd);
	t3_request_buffer(cmd) = (unsigned char *)pSgTrans->new_req_buf;

	new_sg = (struct scatterlist *) t3_request_buffer(cmd);
	old_sg = (struct scatterlist *) pSgTrans->old_req_buf;

	BB_LOCK(&bb->bb_lock, flags);

	if( WRITE_CMD(cmd) ){
		t3_use_sg(cmd) = sg_copy(old_sg, new_sg, pSgTrans->old_sgcnt, BB_MAX_SG, DO_COPY);
	}else	
		t3_use_sg(cmd) = sg_copy(old_sg, new_sg, pSgTrans->old_sgcnt, BB_MAX_SG, NO_COPY);
	BB_UNLOCK(&bb->bb_lock, flags);

        return 0;
}

int bb2sg(Scsi_Cmnd* cmd, bpOSD_COMP_EXTENSION p){
	struct scatterlist *new_sg, *old_sg;
	bb_t	*bb = &p->bb;
	pSGTRANS_T      pSgTrans = NULL;
        int             i;
	unsigned long 	flags;

	bb_msg(BB_DEBUG, "bb2sg CMD: %p, %x\n", cmd, cmd->cmnd[0]);
	//Get this pSgTrans.
	for(i=0; i<BB_PAGES; i++){
		if(bb->SgTransTbl[i].cmd == cmd){
			pSgTrans = &bb->SgTransTbl[i];
			break;
		}
	}	

	if( !pSgTrans ) // does not means lost, but means array from functional --> critical, omit it.
		return 0;

	new_sg = pSgTrans->new_req_buf;
	old_sg = pSgTrans->old_req_buf;

	BB_LOCK(&bb->bb_lock, flags);
	if( READ_CMD(cmd) ){
		sg_copy(new_sg, old_sg, t3_use_sg(cmd), pSgTrans->old_sgcnt, DO_COPY);	
	}else   
		sg_copy(new_sg, old_sg, t3_use_sg(cmd), pSgTrans->old_sgcnt, NO_COPY);
	BB_UNLOCK(&bb->bb_lock, flags);

	t3_use_sg(cmd) = pSgTrans->old_sgcnt;
	t3_request_buffer(cmd) = pSgTrans->old_req_buf;	

	put_sg_translator(p, pSgTrans);

	bb->cmd_reject_threshold--;
}

inline void buffer_assignment(int index, bb_t *bb){
	int start, end, i;

	
	start = index * BB_MAX_SG;
	end = (index+1) * BB_MAX_SG;

	bb->SgTransTbl[index].bbuf = &bb->BBuf[start];
	bb->BBuf[end-1].next = NULL;
}

int __devinit bounce_buffer_init(bpOSD_COMP_EXTENSION p){
	bb_t	*bb = &p->bb;
        unsigned char *VirtAddr;
	dma_addr_t PhyAddr; //dma_handle
        int idx, j;

        VirtAddr = pci_alloc_consistent(NULL, BB_SIZE, &PhyAddr);

        if(!VirtAddr)
                return -ENOMEM;

        for(idx=0;idx<BB_PAGES;idx++){
                bb->BBuf[idx].VirtAddr = VirtAddr + (idx * BB_PAGE_SIZE);
                bb->BBuf[idx].PhyAddr = (unsigned int)PhyAddr + (idx * BB_PAGE_SIZE);
		bb->BBuf[idx].Page = virt_to_page(bb->BBuf[idx].VirtAddr);
                bb->BBuf[idx].next = &bb->BBuf[idx+1];
        }
        bb->BBuf[BB_PAGES-1].next = NULL;

        for(idx=0; idx<BB_CAN_QUEUE; idx++){
		buffer_assignment( idx, bb );
		
		// scatterlist assignment.
		for(j=0;j<BB_MAX_SG;j++){
		#if KERNEL_VERSION_BEFORE(2,6,24)
			bb->SgTransTbl[idx].request_buffer[j].page = (bb->SgTransTbl[idx].bbuf + j)->Page;
		#else
			bb->SgTransTbl[idx].request_buffer[j].page_link = t3_sg_page((bb->SgTransTbl[idx].bbuf + j)); 
		#endif
			bb->SgTransTbl[idx].request_buffer[j].offset = 0;
			bb->SgTransTbl[idx].request_buffer[j].dma_address = (bb->SgTransTbl[idx].bbuf + j)->PhyAddr;
			bb->SgTransTbl[idx].request_buffer[j].length = BB_PAGE_SIZE;
		}

		bb->SgTransTbl[idx].new_req_buf = bb->SgTransTbl[idx].request_buffer;
                bb->SgTransTbl[idx].next = &bb->SgTransTbl[idx+1];
        }
        bb->SgTransTbl[BB_CAN_QUEUE-1].next = NULL;
        bb->pSgTransCurrent = &bb->SgTransTbl[0];

        spin_lock_init(&bb->sg_trans_lock);
        spin_lock_init(&bb->bb_lock);
        return 0;
}

void __devexit bounce_buffer_free(bpOSD_COMP_EXTENSION p){
	bb_t    *bb = &p->bb;
        pci_free_consistent(NULL, BB_SIZE, bb->BBuf[0].VirtAddr, bb->BBuf[0].PhyAddr);
}

#define BOUNCE_BUFFER_INIT(x)   \
        do{     \
		if( bounce_buffer_init(x)<0 ){  \
			x->bb.present = 0;      \
		}else{                          \
			x->bb.present = 1;      \
		}                               \
        }while(0)
#endif

#define BOUNCE_BUFFER_RELEASE(x)	\
	do{	\
		if(x->bb.present)	\
			bounce_buffer_free(x);	\
	}while(0)
